@@ -4,8 +4,8 @@
  * All Rights Reserved
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Affero General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -13,7 +13,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU Affero General Public License
+ * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  */
@@ -642,7 +642,7 @@ int bts_tfi_find_free(const struct gprs_rlcmac_bts *bts, enum gprs_rlcmac_tbf_di
 
 	/* find a TFI that is unused on all PDCH */
 	for (trx = trx_from; trx <= trx_to; trx++) {
-		uint8_t tmp_first_tfi;
+		uint8_t tmp_first_tfi = 0xff; /* make gcc happy */
 		unsigned int tmp_cnt;
 		tmp_cnt = trx_count_free_tfi(&bts->trx[trx], dir, &tmp_first_tfi);
 		if (tmp_cnt > best_cnt) {
@@ -1033,10 +1033,7 @@ int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 		rip->burst_type);
 	bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_UL_TBF);
 	if (plen >= 0) {
-		if (the_pcu->pcu_if_version >= 0x0b)
-			pcu_l1if_tx_agch2(bts, bv, plen, false, GSM_RESERVED_TMSI);
-		else
-			pcu_l1if_tx_agch(bts, bv, plen);
+		pcu_l1if_tx_agch2(bts, bv, plen, false, GSM_RESERVED_TMSI);
 		rc = 0;
 	} else {
 		rc = plen;
@@ -1050,12 +1047,8 @@ send_imm_ass_rej:
 		bv, rip->ra, rip->rfn, rip->burst_type,
 		(uint8_t)osmo_tdef_get(bts->T_defs_bts, 3142, OSMO_TDEF_S, -1));
 	bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_REJ);
-	if (plen >= 0) {
-		if (the_pcu->pcu_if_version >= 0x0b)
-			pcu_l1if_tx_agch2(bts, bv, plen, false, GSM_RESERVED_TMSI);
-		else
-			pcu_l1if_tx_agch(bts, bv, plen);
-	}
+	if (plen >= 0)
+		pcu_l1if_tx_agch2(bts, bv, plen, false, GSM_RESERVED_TMSI);
 	bitvec_free(bv);
 	/* rc was already properly set before goto */
 	return rc;
@@ -1136,24 +1129,20 @@ void bts_snd_dl_ass(struct gprs_rlcmac_bts *bts, const struct gprs_rlcmac_dl_tbf
 	if (plen >= 0) {
 		bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_DL_TBF);
 
-		if (the_pcu->pcu_if_version >= 0x0b) {
-			if (ms_imsi_is_valid(tbf->ms())) {
-				pcu_l1if_tx_pch2(bts, immediate_assignment, plen, true, tbf->imsi(), tbf->tlli());
-			} else {
-				/* During GMM ATTACH REQUEST, the IMSI is not yet known to the PCU or SGSN. (It is
-				 * requested after the GMM ATTACH REQUEST with the GMM IDENTITY REQUEST.) When the PCU
-				 * has to assign a DL TBF but the IMSI is not known, then the IMMEDIATE ASSIGNMENT is
-				 * sent on the AGCH. The reason for this is that without an IMSI we can not calculate
-				 + the paging group, which would be necessary for transmission on PCH. Since the IMSI
-				 * is usually only unknown during the GMM ATTACH REQUEST, we may assume that the MS
-				 * is in non-DRX mode and hence it is listening on all CCCH blocks, including AGCH.
-				 *
-				 * See also: 3gpp TS 44.060, section 5.5.1.5
-				 *           3gpp TS 45.002, section 6.5.3, 6.5.6 */
-				pcu_l1if_tx_agch2(bts, immediate_assignment, plen, true, tbf->tlli());
-			}
+		if (ms_imsi_is_valid(tbf->ms())) {
+			pcu_l1if_tx_pch2(bts, immediate_assignment, plen, true, tbf->imsi(), tbf->tlli());
 		} else {
-			pcu_l1if_tx_pch(bts, immediate_assignment, plen, tbf->imsi());
+			/* During GMM ATTACH REQUEST, the IMSI is not yet known to the PCU or SGSN. (It is
+			 * requested after the GMM ATTACH REQUEST with the GMM IDENTITY REQUEST.) When the PCU
+			 * has to assign a DL TBF but the IMSI is not known, then the IMMEDIATE ASSIGNMENT is
+			 * sent on the AGCH. The reason for this is that without an IMSI we can not calculate
+			 * the paging group, which would be necessary for transmission on PCH. Since the IMSI
+			 * is usually only unknown during the GMM ATTACH REQUEST, we may assume that the MS
+			 * is in non-DRX mode and hence it is listening on all CCCH blocks, including AGCH.
+			 *
+			 * See also: 3gpp TS 44.060, section 5.5.1.5
+			 *           3gpp TS 45.002, section 6.5.3, 6.5.6 */
+			pcu_l1if_tx_agch2(bts, immediate_assignment, plen, true, tbf->tlli());
 		}
 	}
 
