@@ -24,7 +24,7 @@
 #include "pcu_utils.h"
 #include "nacc_fsm.h"
 #include "tbf_ul_ack_fsm.h"
-#include "gprs_ms_storage.h"
+#include "alloc_algo.h"
 
 #include <time.h>
 
@@ -62,30 +62,49 @@ static int64_t now_msec()
 	return (int64_t)(ts.tv_sec) * 1000 + ts.tv_nsec / 1000000;
 }
 
-void gprs_default_cb_ms_idle(struct GprsMs *ms)
-{
-	talloc_free(ms);
-}
+static void ms_becomes_idle(struct GprsMs *ms);
 
-void gprs_default_cb_ms_active(struct GprsMs *ms)
+static int ms_use_cb(struct osmo_use_count_entry *e, int32_t old_use_count, const char *file, int line)
 {
-	/* do nothing */
-}
+	struct GprsMs *ms = e->use_count->talloc_object;
+	int32_t total;
+	int level;
+	char buf[1024];
 
-static struct gpr_ms_callback gprs_default_cb = {
-	.ms_idle = gprs_default_cb_ms_idle,
-	.ms_active = gprs_default_cb_ms_active,
-};
+	if (!e->use)
+		return -EINVAL;
+
+	total = osmo_use_count_total(&ms->use_count);
+
+	if (total == 0
+	    || (total == 1 && old_use_count == 0 && e->count == 1))
+		level = LOGL_INFO;
+	else
+		level = LOGL_DEBUG;
+
+
+	LOGPSRC(DMS, level, file, line, "%s: %s %s: now used by %s\n",
+		ms_name(ms),
+		(e->count - old_use_count) > 0 ? "+" : "-", e->use,
+		(osmo_use_count_to_str_buf(buf, sizeof(buf), &ms->use_count), buf));
+
+	if (e->count < 0)
+		return -ERANGE;
+
+	if (total == 0) {
+		OSMO_ASSERT(ms_is_idle(ms));
+		ms_becomes_idle(ms);
+	}
+	return 0;
+}
 
 static void ms_release_timer_cb(void *data)
 {
 	struct GprsMs *ms = (struct GprsMs *) data;
-	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Release timer expired\n");
+	LOGPMS(ms, DMS, LOGL_INFO, "Release timer expired\n");
 
-	if (ms->timer.data) {
-		ms->timer.data = NULL;
-		ms_unref(ms);
-	}
+	/* Finally free the MS after being idle for a while according to config */
+	talloc_free(ms);
 }
 
 static void ms_llc_timer_cb(void *_ms)
@@ -104,31 +123,36 @@ static void ms_llc_timer_cb(void *_ms)
 }
 
 static int ms_talloc_destructor(struct GprsMs *ms);
-struct GprsMs *ms_alloc(struct gprs_rlcmac_bts *bts, uint32_t tlli)
+struct GprsMs *ms_alloc(struct gprs_rlcmac_bts *bts, const char *use_ref)
 {
 	struct GprsMs *ms = talloc_zero(tall_pcu_ctx, struct GprsMs);
 	OSMO_ASSERT(bts);
 
 	talloc_set_destructor(ms, ms_talloc_destructor);
 
+	llist_add(&ms->list, &bts->ms_list);
+	bts_stat_item_inc(bts, STAT_MS_PRESENT);
+
 	ms->bts = bts;
-	ms->cb = gprs_default_cb;
-	ms->tlli = tlli;
+	ms->tlli = GSM_RESERVED_TMSI;
 	ms->new_ul_tlli = GSM_RESERVED_TMSI;
 	ms->new_dl_tlli = GSM_RESERVED_TMSI;
 	ms->ta = GSM48_TA_INVALID;
 	ms->current_cs_ul = UNKNOWN;
 	ms->current_cs_dl = UNKNOWN;
-	ms->is_idle = true;
-	INIT_LLIST_HEAD(&ms->list);
 	INIT_LLIST_HEAD(&ms->old_tbfs);
 
+	ms->use_count = (struct osmo_use_count){
+		.talloc_object = ms,
+		.use_cb = ms_use_cb,
+	};
+
 	int codel_interval = LLC_CODEL_USE_DEFAULT;
 
-	LOGP(DRLCMAC, LOGL_INFO, "Creating MS object, TLLI = 0x%08x\n", tlli);
+	LOGP(DMS, LOGL_INFO, "Creating MS object\n");
 
 	ms->imsi[0] = '\0';
-	osmo_timer_setup(&ms->timer, ms_release_timer_cb, NULL);
+	osmo_timer_setup(&ms->release_timer, ms_release_timer_cb, ms);
 	llc_queue_init(&ms->llc_queue, ms);
 	memset(&ms->llc_timer, 0, sizeof(ms->llc_timer));
 	osmo_timer_setup(&ms->llc_timer, ms_llc_timer_cb, ms);
@@ -147,6 +171,8 @@ struct GprsMs *ms_alloc(struct gprs_rlcmac_bts *bts, uint32_t tlli)
 	if (!ms->ctrs)
 		goto free_ret;
 
+	if (use_ref)
+		ms_ref(ms, use_ref);
 	return ms;
 free_ret:
 	talloc_free(ms);
@@ -157,25 +183,28 @@ static int ms_talloc_destructor(struct GprsMs *ms)
 {
 	struct llist_item *pos, *tmp;
 
-	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Destroying MS object\n");
+	LOGPMS(ms, DMS, LOGL_INFO, "Destroying MS object\n");
+
+	bts_stat_item_dec(ms->bts, STAT_MS_PRESENT);
+	llist_del(&ms->list);
 
 	ms_set_reserved_slots(ms, NULL, 0, 0);
 
-	osmo_timer_del(&ms->timer);
+	osmo_timer_del(&ms->release_timer);
 
 	if (ms->ul_tbf) {
-		tbf_set_ms(ul_tbf_as_tbf(ms->ul_tbf), NULL);
-		ms->ul_tbf = NULL;
+		tbf_free(ul_tbf_as_tbf(ms->ul_tbf));
+		OSMO_ASSERT(ms->ul_tbf == NULL);
 	}
 
 	if (ms->dl_tbf) {
-		tbf_set_ms(dl_tbf_as_tbf(ms->dl_tbf), NULL);
-		ms->dl_tbf = NULL;
+		tbf_free(dl_tbf_as_tbf(ms->dl_tbf));
+		OSMO_ASSERT(ms->dl_tbf == NULL);
 	}
 
 	llist_for_each_entry_safe(pos, tmp, &ms->old_tbfs, list) {
 		struct gprs_rlcmac_tbf *tbf = (struct gprs_rlcmac_tbf *)pos->entry;
-		tbf_set_ms(tbf, NULL);
+		tbf_free(tbf);
 	}
 
 	llc_queue_clear(&ms->llc_queue, ms->bts);
@@ -186,70 +215,47 @@ static int ms_talloc_destructor(struct GprsMs *ms)
 	return 0;
 }
 
-
-void ms_set_callback(struct GprsMs *ms, struct gpr_ms_callback *cb)
+/* MS has no attached TBFs anymore. */
+static void ms_becomes_idle(struct GprsMs *ms)
 {
-	if (cb)
-		ms->cb = *cb;
-	else
-		ms->cb = gprs_default_cb;
-}
+	unsigned long delay_rel_sec = osmo_tdef_get(ms->bts->pcu->T_defs, -2030, OSMO_TDEF_S, -1);
 
-static void ms_update_status(struct GprsMs *ms)
-{
-	if (ms->ref > 0)
-		return;
+	osmo_gettimeofday(&ms->tv_idle_start, NULL);
 
-	if (ms_is_idle(ms) && !ms->is_idle) {
-		ms->is_idle = true;
-		ms->cb.ms_idle(ms);
-		/* this can be deleted by now, do not access it */
+	ms_set_reserved_slots(ms, NULL, 0, 0);
+	ms->first_common_ts = NULL;
+
+	/* Immediate free():
+	 * Skip delaying free() through release timer if delay is configured to be 0.
+	 * This is useful for synced freed during unit tests.
+	 */
+	if (delay_rel_sec == 0) {
+		talloc_free(ms);
 		return;
 	}
 
-	if (!ms_is_idle(ms) && ms->is_idle) {
-		ms->is_idle = false;
-		ms->cb.ms_active(ms);
+	/* Immediate free():
+	 * Skip delaying free() through release timer if TMSI is not
+	 * known, since those cannot really be reused.
+	 */
+	if (ms_tlli(ms) == GSM_RESERVED_TMSI) {
+		talloc_free(ms);
+		return;
 	}
+
+	LOGPMS(ms, DMS, LOGL_INFO, "Schedule MS release in %lu secs\n", delay_rel_sec);
+	osmo_timer_schedule(&ms->release_timer, delay_rel_sec, 0);
 }
 
-struct GprsMs *ms_ref(struct GprsMs *ms)
+static void ms_becomes_active(struct GprsMs *ms)
 {
-	ms->ref += 1;
-	return ms;
-}
-
-void ms_unref(struct GprsMs *ms)
-{
-	OSMO_ASSERT(ms->ref >= 0);
-	ms->ref -= 1;
-	if (ms->ref == 0)
-		ms_update_status(ms);
-}
-
-static void ms_release_timer_start(struct GprsMs *ms)
-{
-	if (ms->delay == 0)
+	if (!osmo_timer_pending(&ms->release_timer))
 		return;
 
-	LOGPMS(ms, DRLCMAC, LOGL_DEBUG, "Schedule MS release in %u secs\n", ms->delay);
+	LOGPMS(ms, DMS, LOGL_DEBUG, "Cancel scheduled MS release\n");
 
-	if (!ms->timer.data)
-		ms->timer.data = ms_ref(ms);
-
-	osmo_timer_schedule(&ms->timer, ms->delay, 0);
-}
-
-static void ms_release_timer_stop(struct GprsMs *ms)
-{
-	if (!ms->timer.data)
-		return;
-
-	LOGPMS(ms, DRLCMAC, LOGL_DEBUG, "Cancel scheduled MS release\n");
-
-	osmo_timer_del(&ms->timer);
-	ms->timer.data = NULL;
-	ms_unref(ms);
+	timerclear(&ms->tv_idle_start);
+	osmo_timer_del(&ms->release_timer);
 }
 
 void ms_set_mode(struct GprsMs *ms, enum mcs_kind mode)
@@ -303,101 +309,116 @@ void ms_set_mode(struct GprsMs *ms, enum mcs_kind mode)
 	}
 }
 
+/* If a TBF is attached to an MS, it is either in ms->{dl,ul}_tbf or in ms->old_tbfs list */
+static bool ms_tbf_is_attached(const struct GprsMs *ms, const struct gprs_rlcmac_tbf *tbf)
+{
+	const struct llist_item *pos;
+	OSMO_ASSERT(ms);
+	OSMO_ASSERT(tbf);
+	OSMO_ASSERT(tbf_ms(tbf) == ms);
+
+	if (tbf == ul_tbf_as_tbf_const(ms->ul_tbf))
+		return true;
+
+	if (tbf == dl_tbf_as_tbf_const(ms->dl_tbf))
+		return true;
+
+	llist_for_each_entry(pos, &ms->old_tbfs, list) {
+		const struct gprs_rlcmac_tbf *tmp_tbf = (struct gprs_rlcmac_tbf *)pos->entry;
+		if (tmp_tbf == tbf)
+			return true;
+	}
+	return false;
+}
+
 static void ms_attach_ul_tbf(struct GprsMs *ms, struct gprs_rlcmac_ul_tbf *tbf)
 {
-	if (ms->ul_tbf == tbf)
-		return;
-
-	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Attaching UL TBF: %s\n", tbf_name((struct gprs_rlcmac_tbf *)tbf));
-
-	ms_ref(ms);
+	LOGPMS(ms, DMS, LOGL_INFO, "Attaching UL TBF: %s\n", tbf_name((struct gprs_rlcmac_tbf *)tbf));
 
 	if (ms->ul_tbf)
 		llist_add_tail(tbf_ms_list(ul_tbf_as_tbf(ms->ul_tbf)), &ms->old_tbfs);
 
 	ms->ul_tbf = tbf;
 
-	if (tbf)
-		ms_release_timer_stop(ms);
-
-	ms_unref(ms);
+	ms_ref(ms, MS_USE_TBF);
 }
 
 static void ms_attach_dl_tbf(struct GprsMs *ms, struct gprs_rlcmac_dl_tbf *tbf)
 {
-	if (ms->dl_tbf == tbf)
-		return;
-
-	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Attaching DL TBF: %s\n", tbf_name((struct gprs_rlcmac_tbf *)tbf));
-
-	ms_ref(ms);
+	LOGPMS(ms, DMS, LOGL_INFO, "Attaching DL TBF: %s\n", tbf_name((struct gprs_rlcmac_tbf *)tbf));
 
 	if (ms->dl_tbf)
 		llist_add_tail(tbf_ms_list(dl_tbf_as_tbf(ms->dl_tbf)), &ms->old_tbfs);
 
 	ms->dl_tbf = tbf;
 
-	if (tbf)
-		ms_release_timer_stop(ms);
-
-	ms_unref(ms);
+	ms_ref(ms, MS_USE_TBF);
 }
 
 void ms_attach_tbf(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf)
 {
+	OSMO_ASSERT(ms);
+	OSMO_ASSERT(tbf);
+	OSMO_ASSERT(!ms_tbf_is_attached(ms, tbf));
+
 	if (tbf_direction(tbf) == GPRS_RLCMAC_DL_TBF)
 		ms_attach_dl_tbf(ms, tbf_as_dl_tbf(tbf));
 	else
 		ms_attach_ul_tbf(ms, tbf_as_ul_tbf(tbf));
+
+	ms_becomes_active(ms);
 }
 
 void ms_detach_tbf(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf)
 {
-	if (tbf == (struct gprs_rlcmac_tbf *)(ms->ul_tbf)) {
-		ms->ul_tbf = NULL;
-	} else if (tbf == (struct gprs_rlcmac_tbf *)(ms->dl_tbf)) {
-		ms->dl_tbf = NULL;
-	} else {
-		bool found = false;
+	OSMO_ASSERT(tbf_ms(tbf) == ms);
 
-		struct llist_item *pos, *tmp;
-		llist_for_each_entry_safe(pos, tmp, &ms->old_tbfs, list) {
-			struct gprs_rlcmac_tbf *tmp_tbf = (struct gprs_rlcmac_tbf *)pos->entry;
-			if (tmp_tbf == tbf) {
-				llist_del(&pos->list);
-				found = true;
-				break;
-			}
-		}
+	/* In general this should not happen, but it can happen if during TBF
+	 * allocation something fails before tbf->setup() called ms_attach_tbf(). */
+	if (!ms_tbf_is_attached(ms, tbf))
+		return;
 
-		/* Protect against recursive calls via set_ms() */
-		if (!found)
-			return;
-	}
-
-	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Detaching TBF: %s\n",
+	LOGPMS(ms, DMS, LOGL_INFO, "Detaching TBF: %s\n",
 	       tbf_name(tbf));
 
-	if (tbf_ms(tbf) == ms)
-		tbf_set_ms(tbf, NULL);
-
-	if (!ms->dl_tbf && !ms->ul_tbf) {
-		ms_set_reserved_slots(ms, NULL, 0, 0);
-		ms->first_common_ts = NULL;
-
-		if (ms_tlli(ms) != 0)
-			ms_release_timer_start(ms);
+	if (tbf == ul_tbf_as_tbf(ms->ul_tbf)) {
+		ms->ul_tbf = NULL;
+	} else if (tbf == dl_tbf_as_tbf(ms->dl_tbf)) {
+		ms->dl_tbf = NULL;
+	} else {
+		/* We know from ms_tbf_is_attached()==true check above that tbf
+		 * is in ms->old_tbfs, no need to look it up again. */
+		llist_del(tbf_ms_list(tbf));
 	}
 
-	ms_update_status(ms);
+	ms_unref(ms, MS_USE_TBF);
 }
 
-void ms_reset(struct GprsMs *ms)
+/* Cleans up old MS being merged into a new one. Should be called with a
+ * ms_ref() taken to avoid use-after-free.
+ */
+static void ms_reset(struct GprsMs *ms)
 {
-	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Clearing MS object\n");
+	LOGPMS(ms, DMS, LOGL_INFO, "Clearing MS object\n");
+	struct llist_item *pos;
+	struct gprs_rlcmac_tbf *tbf;
 
-	ms_release_timer_stop(ms);
+	tbf = ul_tbf_as_tbf(ms_ul_tbf(ms));
+	if (tbf && !tbf_timers_pending(tbf, T_MAX))
+		tbf_free(tbf);
+	tbf = dl_tbf_as_tbf(ms_dl_tbf(ms));
+	if (tbf && !tbf_timers_pending(tbf, T_MAX))
+		tbf_free(tbf);
 
+	while ((pos = llist_first_entry_or_null(&ms->old_tbfs, struct llist_item, list))) {
+		tbf = (struct gprs_rlcmac_tbf *)pos->entry;
+		if (!tbf_timers_pending(tbf, T_MAX))
+			tbf_free(tbf);
+	}
+
+	/* Flag it with invalid data so that it cannot be looked up anymore and
+	* shows up specially if listed in VTY. Furthermore, it will also trigger
+	* immediate free() when it becomes idle: */
 	ms->tlli = GSM_RESERVED_TMSI;
 	ms->new_dl_tlli = ms->tlli;
 	ms->new_ul_tlli = ms->tlli;
@@ -421,7 +442,7 @@ void ms_update_announced_tlli(struct GprsMs *ms, uint32_t tlli)
 	 * MS object that belongs to that TLLI and if yes make sure one of them
 	 * gets deleted. */
 	if (!ms_check_tlli(ms, tlli))
-		old_ms = ms_store_get_ms(bts_ms_store(ms->bts), tlli, GSM_RESERVED_TMSI, NULL);
+		old_ms = bts_get_ms_by_tlli(ms->bts, tlli, GSM_RESERVED_TMSI);
 
 	ms_set_tlli(ms, tlli);
 
@@ -435,12 +456,14 @@ void ms_update_announced_tlli(struct GprsMs *ms, uint32_t tlli)
 void ms_merge_and_clear_ms(struct GprsMs *ms, struct GprsMs *old_ms)
 {
 	char old_ms_name[128];
+	struct gprs_rlcmac_dl_tbf *dl_tbf;
+
 	OSMO_ASSERT(old_ms != ms);
-	ms_ref(old_ms);
+	ms_ref(old_ms, __func__);
 
 	ms_name_buf(old_ms, old_ms_name, sizeof(old_ms_name));
 
-	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Merge MS: %s\n", old_ms_name);
+	LOGPMS(ms, DMS, LOGL_INFO, "Merge MS: %s\n", old_ms_name);
 
 	if (strlen(ms_imsi(ms)) == 0 && strlen(ms_imsi(old_ms)) != 0)
 		osmo_strlcpy(ms->imsi, ms_imsi(old_ms), sizeof(ms->imsi));
@@ -451,18 +474,18 @@ void ms_merge_and_clear_ms(struct GprsMs *ms, struct GprsMs *old_ms)
 	if (!ms_egprs_ms_class(ms) && ms_egprs_ms_class(old_ms))
 		ms_set_egprs_ms_class(ms, ms_egprs_ms_class(old_ms));
 
+
+	if ((dl_tbf = ms_dl_tbf(old_ms))) {
+		/* Move the last partially/totally unacked LLC PDU back to the LLC queue: */
+		dl_tbf_copy_unacked_pdus_to_llc_queue(dl_tbf);
+	}
+	/* Now merge the old_ms queue into the new one: */
 	llc_queue_move_and_merge(&ms->llc_queue, &old_ms->llc_queue);
 
 	/* Clean up the old MS object */
-	/* TODO: Use timer? */
-	if (ms_ul_tbf(old_ms) && !tbf_timers_pending((struct gprs_rlcmac_tbf *)ms_ul_tbf(old_ms), T_MAX))
-			tbf_free((struct gprs_rlcmac_tbf *)ms_ul_tbf(old_ms));
-	if (ms_dl_tbf(old_ms) && !tbf_timers_pending((struct gprs_rlcmac_tbf *)ms_dl_tbf(old_ms), T_MAX))
-			tbf_free((struct gprs_rlcmac_tbf *)ms_dl_tbf(old_ms));
-
 	ms_reset(old_ms);
 
-	ms_unref(old_ms);
+	ms_unref(old_ms, __func__);
 }
 
 /* Apply changes to the TLLI directly, used interally by functions below: */
@@ -486,7 +509,7 @@ void ms_set_tlli(struct GprsMs *ms, uint32_t tlli)
 		return;
 
 	if (tlli != ms->new_dl_tlli) {
-		LOGP(DRLCMAC, LOGL_INFO,
+		LOGP(DMS, LOGL_INFO,
 			"Modifying MS object, UL TLLI: 0x%08x -> 0x%08x, "
 			"not yet confirmed\n",
 			ms_tlli(ms), tlli);
@@ -494,7 +517,7 @@ void ms_set_tlli(struct GprsMs *ms, uint32_t tlli)
 		return;
 	}
 
-	LOGP(DRLCMAC, LOGL_INFO,
+	LOGP(DMS, LOGL_INFO,
 		"Modifying MS object, TLLI: 0x%08x -> 0x%08x, "
 		"already confirmed partly\n",
 		ms->tlli, tlli);
@@ -512,7 +535,7 @@ bool ms_confirm_tlli(struct GprsMs *ms, uint32_t tlli)
 		/* The MS has not sent a message with the new TLLI, which may
 		 * happen according to the spec [TODO: add reference]. */
 
-		LOGP(DRLCMAC, LOGL_INFO,
+		LOGP(DMS, LOGL_INFO,
 			"The MS object cannot fully confirm an unexpected TLLI: 0x%08x, "
 			"partly confirmed\n", tlli);
 		/* Use the network's idea of TLLI as candidate, this does not
@@ -521,7 +544,7 @@ bool ms_confirm_tlli(struct GprsMs *ms, uint32_t tlli)
 		return false;
 	}
 
-	LOGP(DRLCMAC, LOGL_INFO,
+	LOGP(DMS, LOGL_INFO,
 		"Modifying MS object, TLLI: 0x%08x confirmed\n", tlli);
 
 	ms_apply_tlli_change(ms, tlli);
@@ -532,12 +555,12 @@ bool ms_confirm_tlli(struct GprsMs *ms, uint32_t tlli)
 void ms_set_imsi(struct GprsMs *ms, const char *imsi)
 {
 	if (!imsi) {
-		LOGP(DRLCMAC, LOGL_ERROR, "Expected IMSI!\n");
+		LOGP(DMS, LOGL_ERROR, "Expected IMSI!\n");
 		return;
 	}
 
 	if (imsi[0] && strlen(imsi) < 3) {
-		LOGP(DRLCMAC, LOGL_ERROR, "No valid IMSI '%s'!\n",
+		LOGP(DMS, LOGL_ERROR, "No valid IMSI '%s'!\n",
 			imsi);
 		return;
 	}
@@ -545,11 +568,11 @@ void ms_set_imsi(struct GprsMs *ms, const char *imsi)
 	if (strcmp(imsi, ms->imsi) == 0)
 		return;
 
-	LOGP(DRLCMAC, LOGL_INFO,
+	LOGP(DMS, LOGL_INFO,
 		"Modifying MS object, TLLI = 0x%08x, IMSI '%s' -> '%s'\n",
 		ms_tlli(ms), ms->imsi, imsi);
 
-	struct GprsMs *old_ms = bts_ms_by_imsi(ms->bts, imsi);
+	struct GprsMs *old_ms = bts_get_ms_by_imsi(ms->bts, imsi);
 	/* Check if we are going to store a different MS object with already
 	   existing IMSI. This is probably a bug in code calling this function,
 	   since it should take care of this explicitly */
@@ -558,7 +581,7 @@ void ms_set_imsi(struct GprsMs *ms, const char *imsi)
 		* different IMSI */
 		OSMO_ASSERT(old_ms != ms);
 
-		LOGPMS(ms, DRLCMAC, LOGL_NOTICE,
+		LOGPMS(ms, DMS, LOGL_NOTICE,
 		       "IMSI '%s' was already assigned to another "
 		       "MS object: TLLI = 0x%08x, that IMSI will be removed\n",
 		       imsi, ms_tlli(old_ms));
@@ -577,30 +600,18 @@ void ms_set_imsi(struct GprsMs *ms, const char *imsi)
 		tbf_update_state_fsm_name(dl_tbf_as_tbf(ms->dl_tbf));
 }
 
-uint16_t ms_paging_group(struct GprsMs *ms)
-{
-	uint16_t pgroup;
-	if (!ms_imsi_is_valid(ms))
-		return 0; /* 000 is the special "all paging" group */
-	if ((pgroup = imsi2paging_group(ms_imsi(ms))) > 999) {
-		LOGPMS(ms, DRLCMAC, LOGL_ERROR, "IMSI to paging group failed!\n");
-		return 0;
-	}
-	return pgroup;
-}
-
 void ms_set_ta(struct GprsMs *ms, uint8_t ta_)
 {
 	if (ta_ == ms->ta)
 		return;
 
 	if (gsm48_ta_is_valid(ta_)) {
-		LOGP(DRLCMAC, LOGL_INFO,
+		LOGP(DMS, LOGL_INFO,
 		     "Modifying MS object, TLLI = 0x%08x, TA %d -> %d\n",
 		     ms_tlli(ms), ms->ta, ta_);
 		ms->ta = ta_;
 	} else
-		LOGP(DRLCMAC, LOGL_NOTICE,
+		LOGP(DMS, LOGL_NOTICE,
 		     "MS object, TLLI = 0x%08x, invalid TA %d rejected (old "
 		     "value %d kept)\n", ms_tlli(ms), ta_, ms->ta);
 }
@@ -610,7 +621,7 @@ void ms_set_ms_class(struct GprsMs *ms, uint8_t ms_class_)
 	if (ms_class_ == ms->ms_class)
 		return;
 
-	LOGP(DRLCMAC, LOGL_INFO,
+	LOGP(DMS, LOGL_INFO,
 		"Modifying MS object, TLLI = 0x%08x, MS class %d -> %d\n",
 		ms_tlli(ms), ms->ms_class, ms_class_);
 
@@ -622,14 +633,14 @@ void ms_set_egprs_ms_class(struct GprsMs *ms, uint8_t ms_class_)
 	if (ms_class_ == ms->egprs_ms_class)
 		return;
 
-	LOGP(DRLCMAC, LOGL_INFO,
+	LOGP(DMS, LOGL_INFO,
 		"Modifying MS object, TLLI = 0x%08x, EGPRS MS class %d -> %d\n",
 		ms_tlli(ms), ms->egprs_ms_class, ms_class_);
 
 	ms->egprs_ms_class = ms_class_;
 
 	if (!bts_max_mcs_ul(ms->bts) || !bts_max_mcs_dl(ms->bts)) {
-		LOGPMS(ms, DRLCMAC, LOGL_DEBUG,
+		LOGPMS(ms, DMS, LOGL_DEBUG,
 		       "Avoid enabling EGPRS because use of MCS is disabled: ul=%u dl=%u\n",
 			bts_max_mcs_ul(ms->bts), bts_max_mcs_dl(ms->bts));
 		return;
@@ -643,7 +654,7 @@ void ms_set_egprs_ms_class(struct GprsMs *ms, uint8_t ms_class_)
 	} else {
 		ms_set_mode(ms, EGPRS);
 	}
-	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Enabled EGPRS, mode %s\n", mode_name(ms_mode(ms)));
+	LOGPMS(ms, DMS, LOGL_INFO, "Enabled EGPRS, mode %s\n", mode_name(ms_mode(ms)));
 }
 
 void ms_update_error_rate(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf, int error_rate)
@@ -882,7 +893,7 @@ enum CodingScheme ms_current_cs_dl(const struct GprsMs *ms, enum mcs_kind req_mc
 	}
 
 	if (orig_cs != cs)
-		LOGPMS(ms, DRLCMACDL, LOGL_INFO, "MS (mode=%s) suggests transmitting "
+		LOGPMS(ms, DMS, LOGL_INFO, "MS (mode=%s) suggests transmitting "
 			"DL %s, downgrade to %s in order to match TBF & scheduler requirements\n",
 			mode_name(ms_mode(ms)), mcs_name(orig_cs), mcs_name(cs));
 
@@ -1082,10 +1093,13 @@ static void ms_start_llc_timer(struct GprsMs *ms)
 /* Can we get to send a DL TBF ass to the MS? */
 static bool ms_is_reachable_for_dl_ass(const struct GprsMs *ms)
 {
-	struct gprs_rlcmac_ul_tbf *ul_tbf = ms_ul_tbf(ms);
+	const struct gprs_rlcmac_dl_tbf *dl_tbf = ms_dl_tbf(ms);
+	const struct gprs_rlcmac_ul_tbf *ul_tbf = ms_ul_tbf(ms);
 
-	/* This function assumes it is called when no DL TBF is present */
-	OSMO_ASSERT(!ms_dl_tbf(ms));
+	/* This function assumes it is called when no DL TBF is present, or
+	 * alternatively if it's not really in use by the MS (TBF_ST_WAIT_REUSE_TFI) */
+	OSMO_ASSERT(!dl_tbf ||
+		    tbf_state(dl_tbf_as_tbf_const(dl_tbf)) == TBF_ST_WAIT_REUSE_TFI);
 
 	/* 3GPP TS 44.060 sec 7.1.3.1 Initiation of the Packet resource request procedure:
 	* "Furthermore, the mobile station shall not respond to PACKET DOWNLINK ASSIGNMENT
@@ -1103,7 +1117,8 @@ static bool ms_is_reachable_for_dl_ass(const struct GprsMs *ms)
 	if (!ul_tbf)
 		return true;
 	if (ul_tbf_contention_resolution_done(ul_tbf) &&
-	    !tbf_ul_ack_waiting_cnf_final_ack(ul_tbf))
+	    !tbf_ul_ack_waiting_cnf_final_ack(ul_tbf) &&
+	    tbf_state(ul_tbf_as_tbf_const(ul_tbf)) != TBF_ST_RELEASING)
 		return true;
 
 	return false;
@@ -1118,15 +1133,46 @@ static bool ms_is_reachable_for_dl_ass(const struct GprsMs *ms)
  */
 struct gprs_rlcmac_ul_tbf *ms_new_ul_tbf_assigned_pacch(struct GprsMs *ms, int8_t use_trx)
 {
-	const bool single_slot = false;
 	struct gprs_rlcmac_ul_tbf *ul_tbf;
+	const struct alloc_resources_req req = {
+		.bts = ms->bts,
+		.ms = ms,
+		.direction = GPRS_RLCMAC_UL_TBF,
+		.single = false,
+		.use_trx = use_trx,
+	};
+	struct alloc_resources_res res = {};
+	int rc;
 
-	ul_tbf = ul_tbf_alloc(ms->bts, ms, use_trx, single_slot);
+	rc = the_pcu->alloc_algorithm(&req, &res);
+	if (rc < 0) {
+		LOGPMS(ms, DTBF, LOGL_NOTICE,
+			"Timeslot Allocation failed: trx = %d, single_slot = %d\n",
+			req.use_trx, req.single);
+		bts_do_rate_ctr_inc(ms->bts, CTR_TBF_ALLOC_FAIL);
+		return NULL;
+	}
+
+	ul_tbf = ul_tbf_alloc(ms->bts, ms);
 	if (!ul_tbf) {
-		LOGPMS(ms, DTBF, LOGL_NOTICE, "No PDCH resource\n");
+		LOGPMS(ms, DTBF, LOGL_NOTICE, "ul_tbf_alloc() failed\n");
 		/* Caller will most probably send a Imm Ass Reject after return */
 		return NULL;
 	}
+
+	/* Update MS, really allocate the resources */
+	if (res.reserved_ul_slots != ms_reserved_ul_slots(ms) ||
+	    res.reserved_dl_slots != ms_reserved_dl_slots(ms)) {
+		/* The reserved slots have changed, update the MS */
+		ms_set_reserved_slots(ms, res.trx, res.reserved_ul_slots, res.reserved_dl_slots);
+	}
+	ms_set_first_common_ts(ms, res.first_common_ts);
+
+	/* Apply allocated resources to TBF: */
+	ul_tbf_apply_allocated_resources(ul_tbf, &res);
+
+	ms_attach_tbf(ms, ul_tbf_as_tbf(ul_tbf));
+
 	osmo_fsm_inst_dispatch(tbf_state_fi(ul_tbf_as_tbf(ul_tbf)), TBF_EV_ASSIGN_ADD_PACCH, NULL);
 	/* Contention resolution is considered to be done since TLLI is known in MS */
 	return ul_tbf;
@@ -1136,30 +1182,79 @@ struct gprs_rlcmac_ul_tbf *ms_new_ul_tbf_assigned_pacch(struct GprsMs *ms, int8_
  * packet access", where MS requested only 1 PDCH TS (TS 44.018 Table 9.1.8.1). */
 struct gprs_rlcmac_ul_tbf *ms_new_ul_tbf_assigned_agch(struct GprsMs *ms)
 {
-	const int8_t trx_no = -1;
-	const bool single_slot = true;
 	struct gprs_rlcmac_ul_tbf *ul_tbf;
+	const struct alloc_resources_req req = {
+		.bts = ms->bts,
+		.ms = ms,
+		.direction = GPRS_RLCMAC_UL_TBF,
+		.single = true,
+		.use_trx = -1,
+	};
+	struct alloc_resources_res res = {};
+	int rc;
 
-	ul_tbf = ul_tbf_alloc(ms->bts, ms, trx_no, single_slot);
+	rc = the_pcu->alloc_algorithm(&req, &res);
+	if (rc < 0) {
+		LOGPMS(ms, DTBF, LOGL_NOTICE,
+			"Timeslot Allocation failed: trx = %d, single_slot = %d\n",
+			req.use_trx, req.single);
+		bts_do_rate_ctr_inc(ms->bts, CTR_TBF_ALLOC_FAIL);
+		return NULL;
+	}
+
+	ul_tbf = ul_tbf_alloc(ms->bts, ms);
 	if (!ul_tbf) {
-		LOGP(DTBF, LOGL_NOTICE, "No PDCH resource for Uplink TBF\n");
+		LOGPMS(ms, DTBF, LOGL_NOTICE, "ul_tbf_alloc() failed\n");
 		/* Caller will most probably send a Imm Ass Reject after return */
 		return NULL;
 	}
+
+	/* Update MS, really allocate the resources */
+	if (res.reserved_ul_slots != ms_reserved_ul_slots(ms) ||
+	    res.reserved_dl_slots != ms_reserved_dl_slots(ms)) {
+		/* The reserved slots have changed, update the MS */
+		ms_set_reserved_slots(ms, res.trx, res.reserved_ul_slots, res.reserved_dl_slots);
+	}
+	ms_set_first_common_ts(ms, res.first_common_ts);
+
+	/* Apply allocated resources to TBF: */
+	ul_tbf_apply_allocated_resources(ul_tbf, &res);
+
+	ms_attach_tbf(ms, ul_tbf_as_tbf(ul_tbf));
+
 	osmo_fsm_inst_dispatch(tbf_state_fi(ul_tbf_as_tbf(ul_tbf)), TBF_EV_ASSIGN_ADD_CCCH, NULL);
 	return ul_tbf;
 }
 
 /* Create a temporary dummy TBF to Tx a ImmAssReject if allocating a new one during
  * packet resource Request failed. This is similar as ul_tbf_alloc() but without
- * calling tbf->setup() (in charge of TFI/USF allocation), and reusing resources
+ * calling alloc_algo (in charge of TFI/USF allocation), and reusing resources
  * from Packet Resource Request we received. See TS 44.060 sec 7.1.3.2.1  */
 struct gprs_rlcmac_ul_tbf *ms_new_ul_tbf_rejected_pacch(struct GprsMs *ms, struct gprs_rlcmac_pdch *pdch)
 {
 	struct gprs_rlcmac_ul_tbf *ul_tbf;
-	ul_tbf = ul_tbf_alloc_rejected(ms->bts, ms, pdch);
+	struct alloc_resources_res fake_res = {
+		.trx = pdch->trx,
+		.first_common_ts = pdch,
+		.reserved_ul_slots = 0,
+		.reserved_dl_slots = 0,
+		.ass_slots_mask = 0,
+		.upgrade_to_multislot = false,
+		.tfi = TBF_TFI_UNSET,
+		.usf = {0},
+	};
+	ul_tbf = ul_tbf_alloc(ms->bts, ms);
 	if (!ul_tbf)
 		return NULL;
+
+	/* The only one TS is the common, control TS */
+	ms_set_first_common_ts(ms, pdch);
+
+	/* Apply fake resources to TBF, to attach it to the proper TRX/PDCH: */
+	ul_tbf_apply_allocated_resources(ul_tbf, &fake_res);
+
+	ms_attach_tbf(ms, ul_tbf_as_tbf(ul_tbf));
+
 	osmo_fsm_inst_dispatch(tbf_state_fi(ul_tbf_as_tbf(ul_tbf)), TBF_EV_ASSIGN_ADD_PACCH, NULL);
 	osmo_fsm_inst_dispatch(tbf_ul_ass_fi(ul_tbf_as_tbf(ul_tbf)), TBF_UL_ASS_EV_SCHED_ASS_REJ, NULL);
 
@@ -1173,16 +1268,45 @@ struct gprs_rlcmac_ul_tbf *ms_new_ul_tbf_rejected_pacch(struct GprsMs *ms, struc
 int ms_new_dl_tbf_assigned_on_pacch(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf)
 {
 	OSMO_ASSERT(tbf);
-	const int8_t trx_no = tbf_get_trx(tbf)->trx_no;
-	const bool single_slot = false;
 	struct gprs_rlcmac_dl_tbf *dl_tbf;
+	const struct alloc_resources_req req = {
+		.bts = ms->bts,
+		.ms = ms,
+		.direction = GPRS_RLCMAC_DL_TBF,
+		.single = false,
+		.use_trx = tbf_get_trx(tbf)->trx_no,
+	};
+	struct alloc_resources_res res = {};
+	int rc;
 
-	dl_tbf = dl_tbf_alloc(ms->bts, ms, trx_no, single_slot);
-	if (!dl_tbf) {
-		LOGPMS(ms, DTBF, LOGL_NOTICE, "No PDCH resource\n");
+	rc = the_pcu->alloc_algorithm(&req, &res);
+	if (rc < 0) {
+		LOGPMS(ms, DTBF, LOGL_NOTICE,
+			"Timeslot Allocation failed: trx = %d, single_slot = %d\n",
+			req.use_trx, req.single);
+		bts_do_rate_ctr_inc(ms->bts, CTR_TBF_ALLOC_FAIL);
 		return -EBUSY;
 	}
 
+	dl_tbf = dl_tbf_alloc(ms->bts, ms);
+	if (!dl_tbf) {
+		LOGPMS(ms, DTBF, LOGL_NOTICE, "dl_tbf_alloc() failed\n");
+		return -1;
+	}
+
+	/* Update MS, really allocate the resources */
+	if (res.reserved_ul_slots != ms_reserved_ul_slots(ms) ||
+	    res.reserved_dl_slots != ms_reserved_dl_slots(ms)) {
+		/* The reserved slots have changed, update the MS */
+		ms_set_reserved_slots(ms, res.trx, res.reserved_ul_slots, res.reserved_dl_slots);
+	}
+	ms_set_first_common_ts(ms, res.first_common_ts);
+
+	/* Apply allocated resources to TBF: */
+	dl_tbf_apply_allocated_resources(dl_tbf, &res);
+
+	ms_attach_tbf(ms, dl_tbf_as_tbf(dl_tbf));
+
 	LOGPTBFDL(dl_tbf, LOGL_DEBUG, "[DOWNLINK] START (PACCH)\n");
 	dl_tbf_trigger_ass_on_pacch(dl_tbf, tbf);
 	return 0;
@@ -1193,16 +1317,45 @@ int ms_new_dl_tbf_assigned_on_pacch(struct GprsMs *ms, struct gprs_rlcmac_tbf *t
  */
 int ms_new_dl_tbf_assigned_on_pch(struct GprsMs *ms)
 {
-	const int8_t trx_no = -1;
-	const bool single_slot = true;
 	struct gprs_rlcmac_dl_tbf *dl_tbf;
+	const struct alloc_resources_req req = {
+		.bts = ms->bts,
+		.ms = ms,
+		.direction = GPRS_RLCMAC_DL_TBF,
+		.single = true,
+		.use_trx = -1,
+	};
+	struct alloc_resources_res res = {};
+	int rc;
 
-	dl_tbf = dl_tbf_alloc(ms->bts, ms, trx_no, single_slot);
-	if (!dl_tbf) {
-		LOGPMS(ms, DTBF, LOGL_NOTICE, "No PDCH resource\n");
+	rc = the_pcu->alloc_algorithm(&req, &res);
+	if (rc < 0) {
+		LOGPMS(ms, DTBF, LOGL_NOTICE,
+			"Timeslot Allocation failed: trx = %d, single_slot = %d\n",
+			req.use_trx, req.single);
+		bts_do_rate_ctr_inc(ms->bts, CTR_TBF_ALLOC_FAIL);
 		return -EBUSY;
 	}
 
+	dl_tbf = dl_tbf_alloc(ms->bts, ms);
+	if (!dl_tbf) {
+		LOGPMS(ms, DTBF, LOGL_NOTICE, "dl_tbf_alloc() failed\n");
+		return -1;
+	}
+
+	/* Update MS, really allocate the resources */
+	if (res.reserved_ul_slots != ms_reserved_ul_slots(ms) ||
+	    res.reserved_dl_slots != ms_reserved_dl_slots(ms)) {
+		/* The reserved slots have changed, update the MS */
+		ms_set_reserved_slots(ms, res.trx, res.reserved_ul_slots, res.reserved_dl_slots);
+	}
+	ms_set_first_common_ts(ms, res.first_common_ts);
+
+	/* Apply allocated resources to TBF: */
+	dl_tbf_apply_allocated_resources(dl_tbf, &res);
+
+	ms_attach_tbf(ms, dl_tbf_as_tbf(dl_tbf));
+
 	LOGPTBFDL(dl_tbf, LOGL_DEBUG, "[DOWNLINK] START (PCH)\n");
 	dl_tbf_trigger_ass_on_pch(dl_tbf);
 	return 0;
@@ -1227,20 +1380,31 @@ int ms_append_llc_dl_data(struct GprsMs *ms, uint16_t pdu_delay_csec, const uint
 
 	dl_tbf = ms_dl_tbf(ms);
 	if (dl_tbf) {
-		if (tbf_state(dl_tbf_as_tbf_const(dl_tbf)) == TBF_ST_WAIT_RELEASE) {
-			LOGPTBFDL(dl_tbf, LOGL_DEBUG, "in WAIT RELEASE state (T3193), so reuse TBF\n");
+		switch (tbf_state(dl_tbf_as_tbf_const(dl_tbf))) {
+		case TBF_ST_WAIT_RELEASE:
+			LOGPTBFDL(dl_tbf, LOGL_DEBUG, "in WAIT RELEASE state (T3192), so reuse TBF\n");
 			rc = ms_new_dl_tbf_assigned_on_pacch(ms, dl_tbf_as_tbf(dl_tbf));
+			return rc;
+		case TBF_ST_WAIT_REUSE_TFI:
+			/* According to DL TBF state it should be back to CCCH, let's check UL TBF to have more information. */
+			break;
+		case TBF_ST_RELEASING:
+			/* Something went wrong (T3195), delay for later. */
+		default:
+			/* DL TBF in working status (do nothing)*/
+			return 0;
 		}
-	} else {
-		/* Check if we can create a DL TBF to start sending the enqueued
-		 * data. Otherwise it will be triggered later when it is reachable
-		* again. */
-		if (ms_is_reachable_for_dl_ass(ms)) {
-			if (ms_ul_tbf(ms))
-				rc = ms_new_dl_tbf_assigned_on_pacch(ms, ul_tbf_as_tbf(ms_ul_tbf(ms)));
-			else
-				rc = ms_new_dl_tbf_assigned_on_pch(ms);
-		}
+
+	}
+
+	/* Check if we can create a DL TBF to start sending the enqueued
+	 * data. Otherwise it will be triggered later when it is reachable
+	 * again. */
+	if (ms_is_reachable_for_dl_ass(ms)) {
+		if (ms_ul_tbf(ms))
+			rc = ms_new_dl_tbf_assigned_on_pacch(ms, ul_tbf_as_tbf(ms_ul_tbf(ms)));
+		else
+			rc = ms_new_dl_tbf_assigned_on_pch(ms);
 	}
 	return rc;
 }
