@@ -5,8 +5,8 @@
  * All Rights Reserved
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Affero General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -14,7 +14,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU Affero General Public License
+ * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  */
@@ -27,7 +27,6 @@
 #include <gprs_debug.h>
 #include <coding_scheme.h>
 #include <gprs_ms.h>
-#include <gprs_ms_storage.h>
 #include <pcu_l1_if.h>
 #include <rlc.h>
 #include <sba.h>
@@ -139,7 +138,7 @@ static void pdch_unreserve_all_ms_reserved_slots(struct gprs_rlcmac_pdch *pdch)
 	uint8_t ts_rm_mask = (~(1 << pdch->ts_no));
 	struct gprs_rlcmac_trx *trx = pdch->trx;
 
-	llist_for_each(tmp, bts_ms_list(trx->bts)) {
+	llist_for_each(tmp, &trx->bts->ms_list) {
 		struct GprsMs *ms = llist_entry(tmp, typeof(*ms), list);
 		if (ms->current_trx != trx)
 			continue;
@@ -345,28 +344,18 @@ void gprs_rlcmac_pdch::rcv_control_ack(Packet_Control_Acknowledgement_t *packet,
 
 	poll = pdch_ulc_get_node(ulc, fn);
 	if (!poll || poll->type != PDCH_ULC_NODE_TBF_POLL) {
-		LOGPDCH(this, DRLCMAC, LOGL_NOTICE, "PACKET CONTROL ACK with "
-			"unknown FN=%u TLLI=0x%08x (TRX %d TS %d)\n",
-			fn, tlli, trx_no(), ts_no);
-		ms = bts_ms_by_tlli(bts(), tlli, GSM_RESERVED_TMSI);
+		LOGPDCH(this, DRLCMAC, LOGL_NOTICE, "PACKET CONTROL ACK with unknown FN=%u TLLI=0x%08x\n",
+			fn, tlli);
+		ms = bts_get_ms_by_tlli(bts(), tlli, GSM_RESERVED_TMSI);
 		if (ms)
 			LOGPDCH(this, DRLCMAC, LOGL_NOTICE, "PACKET CONTROL ACK with "
-				"unknown TBF corresponds to MS with IMSI %s, TA %d, "
-				"uTBF (TFI=%d, state=%s), dTBF (TFI=%d, state=%s)\n",
-				ms_imsi(ms), ms_ta(ms),
-				ms_ul_tbf(ms) ? ms_ul_tbf(ms)->tfi() : 0,
-				ms_ul_tbf(ms) ? ms_ul_tbf(ms)->state_name() : "None",
-				ms_dl_tbf(ms) ? ms_dl_tbf(ms)->tfi() : 0,
-				ms_dl_tbf(ms) ? ms_dl_tbf(ms)->state_name() : "None");
+				"unknown TBF corresponds to %s\n", ms_name(ms));
 		return;
 	}
 	OSMO_ASSERT(poll->tbf_poll.poll_tbf);
 	tbf = poll->tbf_poll.poll_tbf;
 	reason = poll->tbf_poll.reason;
 
-	/* Reset N3101 counter: */
-	tbf->n_reset(N3101);
-
 	ms_update_announced_tlli(tbf->ms(), tlli);
 	/* Gather MS from TBF again, since it may be NULL or may have been merged during ms_update_announced_tlli */
 	ms = tbf->ms();
@@ -402,8 +391,8 @@ void gprs_rlcmac_pdch::rcv_control_ack(Packet_Control_Acknowledgement_t *packet,
 		}
 		LOGPTBF(tbf, LOGL_DEBUG, "[DOWNLINK] UPLINK ASSIGNED\n");
 		pdch_ulc_release_fn(ulc, fn);
-		/* reset N3105 */
-		tbf->n_reset(N3105);
+		if (tbf_direction(tbf) == GPRS_RLCMAC_DL_TBF)
+			tbf->n_reset(N3105);
 		osmo_fsm_inst_dispatch(tbf->ul_ass_fsm.fi, TBF_UL_ASS_EV_RX_ASS_CTRL_ACK, NULL);
 
 		new_tbf = ms_ul_tbf(ms);
@@ -412,9 +401,6 @@ void gprs_rlcmac_pdch::rcv_control_ack(Packet_Control_Acknowledgement_t *packet,
 				"TBF is gone TLLI=0x%08x\n", tlli);
 			return;
 		}
-		if (tbf->state_is(TBF_ST_WAIT_RELEASE) &&
-				tbf->direction == new_tbf->direction)
-			tbf_free(tbf);
 
 		osmo_fsm_inst_dispatch(new_tbf->state_fi, TBF_EV_ASSIGN_ACK_PACCH, NULL);
 		/* there might be LLC packets waiting in the queue, but the DL
@@ -433,8 +419,8 @@ void gprs_rlcmac_pdch::rcv_control_ack(Packet_Control_Acknowledgement_t *packet,
 		}
 		LOGPTBF(tbf, LOGL_DEBUG, "[UPLINK] DOWNLINK ASSIGNED\n");
 		pdch_ulc_release_fn(ulc, fn);
-		/* reset N3105 */
-		tbf->n_reset(N3105);
+		if (tbf_direction(tbf) == GPRS_RLCMAC_DL_TBF)
+			tbf->n_reset(N3105);
 		osmo_fsm_inst_dispatch(tbf->dl_ass_fsm.fi, TBF_DL_ASS_EV_RX_ASS_CTRL_ACK, NULL);
 
 		new_tbf = ms_dl_tbf(ms);
@@ -443,8 +429,9 @@ void gprs_rlcmac_pdch::rcv_control_ack(Packet_Control_Acknowledgement_t *packet,
 				"TBF is gone TLLI=0x%08x\n", tlli);
 			return;
 		}
-		if (tbf->state_is(TBF_ST_WAIT_RELEASE) &&
-				tbf->direction == new_tbf->direction)
+		if ((tbf->state_is(TBF_ST_WAIT_RELEASE) ||
+		     tbf->state_is(TBF_ST_WAIT_REUSE_TFI)) &&
+		    tbf->direction == new_tbf->direction)
 			tbf_free(tbf);
 
 		osmo_fsm_inst_dispatch(new_tbf->state_fi, TBF_EV_ASSIGN_ACK_PACCH, NULL);
@@ -507,9 +494,6 @@ void gprs_rlcmac_pdch::rcv_control_dl_ack_nack(Packet_Downlink_Ack_Nack_t *ack_n
 		return;
 	}
 
-	/* Reset N3101 counter: */
-	tbf->n_reset(N3101);
-
 	pdch_ulc_release_fn(ulc, fn);
 
 	LOGPTBF(tbf, LOGL_DEBUG, "RX: [PCU <- BTS] Packet Downlink Ack/Nack\n");
@@ -577,8 +561,6 @@ void gprs_rlcmac_pdch::rcv_control_egprs_dl_ack_nack(EGPRS_PD_AckNack_t *ack_nac
 		return;
 	}
 
-	/* Reset N3101 counter: */
-	tbf->n_reset(N3101);
 	pdch_ulc_release_fn(ulc, fn);
 
 	LOGPTBF(tbf, LOGL_DEBUG,
@@ -664,9 +646,11 @@ void gprs_rlcmac_pdch::rcv_resource_request(Packet_Resource_Request_t *request,
 	/* First gather MS from TLLI/DL_TFI/UL_TFI:*/
 	if (request->ID.UnionType) { /* ID_TYPE = TLLI */
 		uint32_t tlli = request->ID.u.TLLI;
-		ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+		ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 		if (!ms) {
-			ms = bts_alloc_ms(bts, 0, 0); /* ms class updated later */
+			/* A reference is already hold immediately below in all cases, no
+			 * need to hold and extra one, pass use_ref=NULL: */
+			ms = ms_alloc(bts, NULL);
 			ms_set_tlli(ms, tlli);
 		}
 	} else if (request->ID.u.Global_TFI.UnionType) { /* ID_TYPE = DL_TFI */
@@ -676,8 +660,6 @@ void gprs_rlcmac_pdch::rcv_resource_request(Packet_Resource_Request_t *request,
 			LOGP(DRLCMAC, LOGL_NOTICE, "PACKET RESOURCE REQ unknown downlink TFI=%d\n", tfi);
 			return;
 		}
-		/* Reset N3101 counter: */
-		dl_tbf->n_reset(N3101);
 		ms = tbf_ms(dl_tbf_as_tbf(dl_tbf));
 		dl_tbf = NULL;
 	} else { /* ID_TYPE = UL_TFI */
@@ -687,8 +669,6 @@ void gprs_rlcmac_pdch::rcv_resource_request(Packet_Resource_Request_t *request,
 			LOGP(DRLCMAC, LOGL_NOTICE, "PACKET RESOURCE REQ unknown uplink TFI=%d\n", tfi);
 			return;
 		}
-		/* Reset N3101 counter: */
-		ul_tbf->n_reset(N3101);
 		ms = tbf_ms(ul_tbf_as_tbf(ul_tbf));
 		ul_tbf = NULL;
 	}
@@ -698,7 +678,7 @@ void gprs_rlcmac_pdch::rcv_resource_request(Packet_Resource_Request_t *request,
 	 * temporarily, in order to avoid it being freed if we free any of its
 	 * resources (TBF). */
 	OSMO_ASSERT(ms);
-	ms_ref(ms);
+	ms_ref(ms, __func__);
 
 
 	switch (item->type) {
@@ -754,7 +734,6 @@ void gprs_rlcmac_pdch::rcv_resource_request(Packet_Resource_Request_t *request,
 		 * contention resolution is done:
 		 */
 		if ((dl_tbf = ms_dl_tbf(ms))) {
-			/* Get rid of previous finished UL TBF before providing a new one */
 			LOGPTBFDL(dl_tbf, LOGL_NOTICE,
 				  "Got PACKET RESOURCE REQ while DL-TBF pending, killing it\n");
 			tbf_free(dl_tbf);
@@ -844,7 +823,7 @@ void gprs_rlcmac_pdch::rcv_resource_request(Packet_Resource_Request_t *request,
 	/* schedule uplink assignment */
 	osmo_fsm_inst_dispatch(new_ul_tbf->ul_ass_fsm.fi, TBF_UL_ASS_EV_SCHED_ASS, NULL);
 return_unref:
-	ms_unref(ms);
+	ms_unref(ms, __func__);
 	return;
 }
 
@@ -853,13 +832,16 @@ void gprs_rlcmac_pdch::rcv_measurement_report(Packet_Measurement_Report_t *repor
 	struct gprs_rlcmac_sba *sba;
 	struct pdch_ulc_node *poll;
 	GprsMs *ms;
+	bool ms_allocated = false;
 
-	ms = bts_ms_by_tlli(bts(), report->TLLI, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts(), report->TLLI, GSM_RESERVED_TMSI);
 	if (!ms) {
 		LOGPDCH(this, DRLCMAC, LOGL_NOTICE, "MS send measurement "
 			"but TLLI 0x%08x is unknown\n", report->TLLI);
-		ms = bts_alloc_ms(bts(), 0, 0);
+		ms = ms_alloc(bts(), __func__);
 		ms_set_tlli(ms, report->TLLI);
+		/* Track that we need to free ref at the end: */
+		ms_allocated = true;
 	}
 	if ((poll = pdch_ulc_get_node(ulc, fn))) {
 		switch (poll->type) {
@@ -880,6 +862,8 @@ void gprs_rlcmac_pdch::rcv_measurement_report(Packet_Measurement_Report_t *repor
 		}
 	}
 	gprs_rlcmac_meas_rep(ms, report);
+	if (ms_allocated)
+		ms_unref(ms, __func__);
 }
 
 void gprs_rlcmac_pdch::rcv_cell_change_notification(Packet_Cell_Change_Notification_t *notif,
@@ -980,9 +964,6 @@ free_ret:
 int gprs_rlcmac_pdch::rcv_block(uint8_t *data, uint8_t len, uint32_t fn,
 	struct pcu_l1_meas *meas)
 {
-	/* First of all, update TDMA clock: */
-	bts_set_current_frame_number(trx->bts, fn);
-
 	/* No successfully decoded UL block was received during this FN: */
 	if (len == 0)
 		return 0;
